#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2023, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_rosbag2csv
# just install "pip install cnspy-rosbag2csv"
########################################################################################################################
import math
from plistlib import Data

from docutils.nodes import topic

import cnspy_numpy_utils
import rosbag
import time
import os
import argparse
import yaml
import pandas as pandas
import csv
from tqdm import tqdm
import numpy as np
from numpy import linalg as LA, number
from cnspy_numpy_utils.numpy_statistics import *
from matplotlib import pyplot as plt
from cnspy_trajectory.PlotLineStyle import PlotLineStyle


class DataframeCfg:

    verbose = False
    label_timestamp = 't'
    label_ID1 = 'UWB_ID1'
    label_ID2 = 'UWB_ID2'
    label_range = 'range_raw'
    def __init__(self, label_timestamp='t',
                 label_ID1='UWB_ID1',
                 label_ID2='UWB_ID2', label_range = 'range_raw'):
        self.label_timestamp = label_timestamp
        self.label_ID1 = label_ID1
        self.label_ID2 = label_ID2
        self.label_range = label_range

class StaticBiasAnalysis:
    dict_err_ij = dict()

    def __init__(self, csv_fn, cfg_fn, verbose=False, df_cfg=DataframeCfg()):
        if not os.path.isfile(csv_fn):
            print("StaticBiasAnalysis: could not find file: %s" % csv_fn)
            return False
        cfg_fn = os.path.abspath(cfg_fn)
        if not os.path.isfile(cfg_fn):
            print("StaticBiasAnalysis: could not find file: %s" % cfg_fn)
            return False
        if verbose:
            print("StaticBiasAnalysis:")
            print("* csv_fn in name: " + str(csv_fn))
            print("* cfg_fn YAML file: \t " + str(cfg_fn))

        dict_cfg = None
        with open(cfg_fn, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)

            if "abs_anchor_positions" not in dict_cfg:
                print("[abs_anchor_positions] does not exist in fn=" + cfg_fn)
                return False
            print("Read successful")

        if verbose:
            print("configuration contains:")
            print("* abs_anchor_positions:" + str(dict_cfg["abs_anchor_positions"]))

        df = pandas.read_csv(csv_fn, sep='\s+|\,', comment='#', engine='python')

        if verbose:
            print("num measurements loaded:" + str(df.size))

        # Iterate all rows using DataFrame.iterrows()
        self.dict_err_ij = dict()
        for index, row in df.iterrows():
            ID1 = row[df_cfg.label_ID1]
            ID2 = row[df_cfg.label_ID2]
            range = row[df_cfg.label_range]
            t = row[df_cfg.label_timestamp]

            if range < 0 or range > 150.0:
                continue

            if ID2 not in dict_cfg["abs_anchor_positions"]:
                #if verbose:
                    #print("ID2=" + str(ID2) + " not in abs_anchor_positions")
                    #print(str(row))
                continue
            if ID1 not in dict_cfg["abs_anchor_positions"]:
                #if verbose:
                    #print("ID1=" + str(ID1) + " not in abs_anchor_positions")
                    #print(str(row))
                continue

            if ID1 not in self.dict_err_ij:
                self.dict_err_ij[ID1] = dict()
            if ID2 not in self.dict_err_ij[ID1]:
                self.dict_err_ij[ID1][ID2] = list()

            t_GA1 = np.array(dict_cfg["abs_anchor_positions"][ID1])
            t_GA2 = np.array(dict_cfg["abs_anchor_positions"][ID2])
            d_A1A2 = LA.norm(t_GA1 - t_GA2)
            self.dict_err_ij[ID1][ID2].append(range - d_A1A2)
        pass


    def plot_ranges(self, ID1, ID2, cfg_dpi=200, cfg_title="ranges", sorted=False, fig=None, ax=None,
                    color='r', label='err',
                    ls=PlotLineStyle(linestyle='-'),
                    save_fn="", result_dir="."):
        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg_dpi))
        if ax is None:
            ax = fig.add_subplot(111)
        if cfg_title:
            ax.set_title(cfg_title)

        r_err_arr = self.dict_err_ij[ID1][ID2]
        x_arr = range(len(r_err_arr))
        StaticBiasAnalysis.ax_plot_n_dim(ax, x_arr, r_err_arr,
                                      colors=color, labels=label, ls=ls)
        ax.grid(b=True)
        ax.set_ylabel('range')
        ax.set_xlabel('range sorted index')
        StaticBiasAnalysis.show_save_figure(fig=fig, result_dir=result_dir, save_fn=save_fn, show=False)
        pass

    def plot_error_histogram(self, ID1, ID2,
                             cfg_dpi=200, fig=None, ax=None,
                             cfg_title="histogram",
                             print_stats=True):
        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg_dpi))
        if ax is None:
            ax = fig.add_subplot(111)
        if cfg_title:
            ax.set_title(cfg_title)

        r_err_arr = self.dict_err_ij[ID1][ID2]
        x_arr = range(len(r_err_arr))
        num_bins = 50
        stat = numpy_statistics(vNumpy=np.squeeze(np.asarray(r_err_arr)))
        if print_stats:
            cnspy_numpy_utils.numpy_statistics.print_statistics(stat, desc=str(ID1) + "-" + str(ID2) + " error")
            pass
        n, bins, patches = ax.hist(r_err_arr, num_bins)

        # add a 'best fit' line
        sigma = stat['std']
        mu = stat['mean']
        y = ((1 / (np.sqrt(2 * np.pi) * sigma)) *
             np.exp(-0.5 * (1 / sigma * (bins - mu)) ** 2))
        ax.plot(bins, y, '--')
        ax.set_ylabel('Probability density')
        ax.set_title(r'Hist: ' + cfg_title + ': $\mu$=' + str(round(mu, 3)) + ', $\sigma$=' + str(round(sigma, 3)))

        pass

    def plot_histograms(self, cfg_dpi=200, result_dir="."):
        plt.style.use('ggplot')

        SMALL_SIZE = 6
        MEDIUM_SIZE = 7
        BIGGER_SIZE = 8

        plt.rc('font', size=SMALL_SIZE)  # controls default text sizes
        plt.rc('axes', titlesize=SMALL_SIZE)  # fontsize of the axes title
        plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
        plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

        for ID1, dict_err_j in self.dict_err_ij.items():
            fig = plt.figure(figsize=(20, 20), dpi=int(cfg_dpi))
            idx = 1
            n = len(dict_err_j)
            sqrt_n = math.ceil(math.sqrt(n))
            n_rows = sqrt_n
            n_cols = sqrt_n
            fig.suptitle('Histograms of ID=' +str(ID1), fontsize=16)
            for ID2, list_err in dict_err_j.items():
                cfg_title = str(str(ID1) + "-" + str(ID2))
                ax = fig.add_subplot(n_rows, n_cols, idx)
                self.plot_error_histogram(ID1=ID1, ID2=ID2, cfg_title=cfg_title,fig=fig, ax=ax)
                idx += 1

            StaticBiasAnalysis.show_save_figure(fig=fig, result_dir=result_dir, save_fn=str("Histograms_ID" + str(ID1)), show=False)
            fig.tight_layout()
            plt.show
            pass

    @staticmethod
    def ax_plot_n_dim(ax, x_linespace, values,
                      colors=['r', 'g', 'b'],
                      labels=['x', 'y', 'z'], ls=PlotLineStyle()):
        assert len(colors) == len(labels)
        if len(colors) > 1:
            assert len(colors) == values.shape[1]
            for i in range(len(colors)):
                ax.plot(x_linespace, values[:, i],
                        color=colors[i], label=labels[i], linestyle=ls.linestyle, linewidth=ls.linewidth,
                        marker=ls.marker)
        else:
            ax.plot(x_linespace, values, color=colors[0], label=labels[0], linestyle=ls.linestyle,
                    linewidth=ls.linewidth)

    @staticmethod
    def ax_x_linespace(ax, ts=None, dist_vec=None, relative_time=True, plot_type=None, x_label_prefix=''):
        if plot_type == "plot_2D_over_dist":
            x_arr = dist_vec
            ax.set_xlabel(x_label_prefix + 'distance [m]')
        else:
            x_arr = ts
            if relative_time:
                x_arr = x_arr - x_arr[0]
                ax.set_xlabel(x_label_prefix + 'rel. time [sec]')
            else:
                ax.set_xlabel(x_label_prefix + 'time [sec]')
        return x_arr

    @staticmethod
    def show_save_figure(fig, save_fn="", result_dir=".", dpi=200, show=True, close_figure=False):
        assert (isinstance(fig, plt.Figure))
        plt.pause(0.01)
        plt.draw()
        plt.pause(0.01)
        if save_fn:
            if not os.path.exists(result_dir):
                os.makedirs(result_dir)

            filename = os.path.join(result_dir, save_fn)
            print("save to file: " + filename)
            plt.savefig(filename, dpi=int(dpi))
        if show:
            plt.show()
        if close_figure:
            plt.close(fig)


def main():
    # example: ROSBag_Pose2Ranges.py --bagfile ../test/sample_data//uwb_calib_a01_2023-08-31-21-05-46.bag --topic /d01/mavros/vision_pose/pose --cfg ../test/sample_data/config.yaml --verbose
    parser = argparse.ArgumentParser(
        description='CSV_StaticBiasAnalysis: extract a given pose topic and compute ranges to N abs_anchor_positions and M rel_tag_positions, which is stored into a CSV file')
    parser.add_argument('--csv_fn', help='input bag file', required=True)
    parser.add_argument('--cfg_fn',
                        help='YAML configuration file describing the setup: {abs_anchor_positions}',
                        default="config.yaml", required=True)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--label_timestamp', help='timestamp label in CSV', default='t')
    parser.add_argument('--label_range', help='range label in CSV', default='range_raw')
    parser.add_argument('--label_ID1', help='ID1 label in CSV', default='UWB_ID1')
    parser.add_argument('--label_ID2', help='ID2 label in CSV', default='UWB_ID2')
    tp_start = time.time()
    args = parser.parse_args()

    df_cfg = DataframeCfg(label_timestamp=args.label_timestamp,
                             label_ID1=args.label_ID1,
                             label_ID2=args.label_ID2,
                             label_range = args.label_range)

    sba = StaticBiasAnalysis(csv_fn=args.csv_fn, cfg_fn=args.cfg_fn, verbose=args.verbose, df_cfg=df_cfg)
    sba.plot_histograms()
    print(" ")
    print("finished after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass




#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2024, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_rosbag2csv
# just install "pip install cnspy-rosbag2csv"
########################################################################################################################
import yaml
from tqdm import tqdm
import numpy as np
from spatialmath import UnitQuaternion, SE3

from cnspy_trajectory.HistoryBuffer import HistoryBuffer, get_key_from_value


class ROSBag_Pose:
    @staticmethod
    def extract_pose(bag, num_messages, topic_pose_body, round_decimals=6, T_BODY_SENSOR=None) -> HistoryBuffer:
        """
        Parameters
        ----------
        bag: file handle to the rosbag
        num_messages: number of messages contained in the bag file
        topic_pose_body: topic name of body pose
        round_decimals (optional): integer > 1
        T_BODY_SENSOR(optional): SE3 pose from Body to Sensor

        Returns HistoryBuffer
        -------
        """
        if round_decimals < 1:
            round_decimals = 1

        hist_poses = dict()
        ## extract the desired topics from the BAG file
        try:  # else already exists
            print("ROSBag_Pose.extract(): extracting pose...")
            cnt_poses = 0
            for topic, msg, t in tqdm(bag.read_messages(), total=num_messages, unit="msgs"):
                if topic == topic_pose_body:
                    T_GLOBAL_BODY = None
                    if hasattr(msg, 'header') and hasattr(msg, 'pose') and hasattr(msg, 'twist'):  # nav_msgs/Odometry Message
                        p = np.array([msg.pose.pose.position.x, msg.pose.pose.position.y, msg.pose.pose.position.z])
                        q_GB = [msg.pose.pose.orientation.w, msg.pose.pose.orientation.x, msg.pose.pose.orientation.y,
                                msg.pose.pose.orientation.z]

                        q = UnitQuaternion(q_GB).unit()
                        T_GLOBAL_BODY = SE3.Rt(q.R, p, check=True)
                        pass
                    elif hasattr(msg, 'header') and hasattr(msg, 'pose'):  # POSE_STAMPED
                        t = np.array([msg.pose.position.x, msg.pose.position.y, msg.pose.position.z])
                        q_GB = [msg.pose.orientation.w, msg.pose.orientation.x, msg.pose.orientation.y,
                                msg.pose.orientation.z]

                        q = UnitQuaternion(q_GB, norm=True)
                        T_GLOBAL_BODY = SE3.Rt(q.R, t, check=True)
                        pass
                    elif hasattr(msg, 'header') and hasattr(msg, 'transform'):
                        t = np.array(
                            [msg.transform.translation.x, msg.transform.translation.y, msg.transform.translation.z])
                        q_GB = [msg.transform.rotation.w, msg.transform.rotation.x, msg.transform.rotation.y,
                                msg.transform.rotation.z]
                        q = UnitQuaternion(q_GB, norm=True)
                        T_GLOBAL_BODY = SE3.Rt(q.R, t, check=True)
                    else:
                        print("\nROSBag_Pose.extract(): unsupported message " + str(msg))
                        continue

                    if T_GLOBAL_BODY is not None:
                        timestamp = round(msg.header.stamp.to_sec(), round_decimals)
                        if T_BODY_SENSOR and isinstance(T_BODY_SENSOR, SE3):
                            hist_poses[timestamp] = T_GLOBAL_BODY * T_BODY_SENSOR
                        else:
                            hist_poses[timestamp] = T_GLOBAL_BODY
                        cnt_poses = cnt_poses + 1
                        pass
                pass

            if cnt_poses == 0:
                print("\nROSBag_Pose.extract(): no poses obtained!")
                return None
            else:
                print("\nROSBag_Pose.extract(): poses extracted: " + str(cnt_poses))
                pass
        except AssertionError as error:
            print(error)
            print(
                "ROSBag_Pose.extract(): Unexpected error while reading the bag file!\n * try: $ rosbag fix <bagfile> <fixed>")
            return None
        return HistoryBuffer(dict_t=hist_poses)

    @staticmethod
    def extract_poses(bag, dict_topic_pose_body, dict_senor_topic_pose, num_messages=None, round_decimals=6, dict_T_BODY_SENSOR=None) -> dict:
        """
        Parameters
        ----------
        bag: file handle to the rosbag
        num_messages: number of messages contained in the bag file
        dict_topic_pose_body: topic names of Body poses
        dict_senor_topic_pose: topic names for the final Sensor pose history
        round_decimals (optional): integer > 1
        dict_T_BODY_SENSOR (optional): SE3 pose from Body to Sensor: T_GLOBAL_SENSOR(t) = T_GLOBAL_BODY(t) * T_BODY_SENSOR

        Returns  a dictionary of HistoryBuffer: dict<senor_topic, HistoryBuffer<T_GLOBAL_SENSOR(t)>>
        -------
        """
        if num_messages is None:
            info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)
            num_messages = info_dict['messages']
        if round_decimals < 1:
            round_decimals = 1


        dict_hist_poses = dict()  # map<topic<timestamp, SE3>>
        for id_, topic_ in dict_senor_topic_pose.items():
            dict_hist_poses[topic_] = dict()

        try:  # else already exists
            print("ROSBag_Pose: extracting poses...")
            cnt_poses = 0
            for topic_, msg, t in tqdm(bag.read_messages(), total=num_messages, unit="msgs"):
                if topic_ in dict_topic_pose_body.values():

                    id_topic = get_key_from_value(dict_topic_pose_body, topic_)
                    topic_sensor = dict_senor_topic_pose[id_topic]
                    T_GLOBAL_BODY = None
                    if hasattr(msg, 'header') and hasattr(msg, 'pose') and hasattr(msg, 'twist'):  # nav_msgs/Odometry Message
                        p = np.array([msg.pose.pose.position.x, msg.pose.pose.position.y, msg.pose.pose.position.z])
                        q_GB = [msg.pose.pose.orientation.w, msg.pose.pose.orientation.x, msg.pose.pose.orientation.y,
                                msg.pose.pose.orientation.z]

                        q = UnitQuaternion(q_GB).unit()
                        T_GLOBAL_BODY = SE3.Rt(q.R, p, check=True)
                        pass
                    elif hasattr(msg, 'header') and hasattr(msg, 'pose'):  # POSE_STAMPED
                        p = np.array([msg.pose.position.x, msg.pose.position.y, msg.pose.position.z])
                        q_GB = [msg.pose.orientation.w, msg.pose.orientation.x, msg.pose.orientation.y,
                                msg.pose.orientation.z]

                        q = UnitQuaternion(q_GB).unit()
                        T_GLOBAL_BODY = SE3.Rt(q.R, p, check=True)
                        pass
                    elif hasattr(msg, 'header') and hasattr(msg, 'transform'):
                        p = np.array(
                            [msg.transform.translation.x, msg.transform.translation.y, msg.transform.translation.z])
                        q_GB = [msg.transform.rotation.w, msg.transform.rotation.x, msg.transform.rotation.y,
                                msg.transform.rotation.z]
                        q = UnitQuaternion(q_GB).unit()
                        T_GLOBAL_BODY = SE3.Rt(q.R, p, check=True)
                    else:
                        print("\nROSBag_Pose: unsupported message " + str(msg))
                        continue

                    if T_GLOBAL_BODY is not None:
                        timestamp = round(msg.header.stamp.to_sec(), round_decimals)

                        if dict_T_BODY_SENSOR and topic_sensor in dict_T_BODY_SENSOR.keys():
                            T_BODY_SENSOR = dict_T_BODY_SENSOR[topic_sensor]
                            dict_hist_poses[topic_sensor][timestamp] = T_GLOBAL_BODY * T_BODY_SENSOR
                        else:
                            dict_hist_poses[topic_sensor][timestamp] = T_GLOBAL_BODY
                        cnt_poses = cnt_poses + 1
                        pass
                pass

            if cnt_poses == 0:
                print("\nROSBag_Pose: no poses obtained!")
                return None
            else:
                print("\nROSBag_Pose: poses extracted: " + str(cnt_poses))

        except AssertionError as error:
            print(error)
            print(
                "ROSBag_Pose: Unexpected error while reading the bag file!\n * try: $ rosbag fix <bagfile> <fixed>")
            return None

        ## convert poses to History only if poses exist in bag file.
        dict_history = dict() # map<topic, history>
        for topic_sensor, hist_poses in dict_hist_poses.items():
            if len(hist_poses):
                dict_history[topic_sensor] = HistoryBuffer(dict_t=hist_poses)

        return dict_history



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################

import os
from cnspy_ranging_evaluation.RangeEvaluation import *
from cnspy_ranging_evaluation.TWR_ROSbag2CSV import *
from cnspy_ranging_evaluation.ROSBag_TrueRanges import *

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')
RES_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_results')

class RangeEvaluationTrail:

    @staticmethod
    def test_eval():
        sequence = 'T1_A3_loiter_2m_2023-08-31-20-58-20'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        topic_list=['/d01/ranging', '/a01/ranging', '/a02/ranging', '/a03/ranging']
        fn_meas_ranges = str(RES_DATA_DIR + '/' + sequence + '-all-meas-ranges.csv')
        fn_gt_ranges = str(RES_DATA_DIR + '/' + sequence + '-all-true-ranges.csv')


        # 1) extract all measurements to CSV
        if not os.path.isfile(fn_meas_ranges):
            res = TWR_ROSbag2CSV.extract_to_one(bagfile_name=bagfile_in,
                                                    topic_list=topic_list,
                                                    fn=fn_meas_ranges,
                                                    result_dir=RES_DATA_DIR,
                                                    verbose=True)



        bagfile_out = str(RES_DATA_DIR + '/' + sequence + '-true-ranges.bag')

        # 2) create a clean bag file
        if not os.path.isfile(bagfile_out):
            res = ROSbag_TrueRanges.extract(bagfile_in_name=bagfile_in,
                                            bagfile_out_name=bagfile_out,
                                            topic_pose='/d01/mavros/vision_pose/pose',
                                            cfg=cfg_file,
                                            verbose=True,
                                            stddev_range=0.001)

        if not os.path.isfile(fn_gt_ranges):
            # 3) extract all measurements from the clean bagfile
            res = TWR_ROSbag2CSV.extract_to_one(bagfile_name=bagfile_out,
                                                    topic_list=topic_list,
                                                    fn=fn_gt_ranges,
                                                    result_dir=RES_DATA_DIR,
                                                    verbose=True)


        # 4) evaluate the ranges
        cfg = AssociateRangesCfg(ID1=None,
                                 ID2=None,
                                 relative_timestamps=False,
                                 max_difference=0.03,
                                 subsample=0,
                                 verbose=True,
                                 remove_outliers=False,
                                 max_range=30,
                                 range_error_val=0,
                                 label_timestamp='t',
                                 label_ID1='UWB_ID1',
                                 label_ID2='UWB_ID2',
                                 label_range='range_raw')

        eval = RangeEvaluation(fn_gt=fn_gt_ranges,
                               fn_est=fn_meas_ranges,
                               UWB_ID1_arr=[0,1,2,3],
                               UWB_ID2_arr=[0,1,2,3],
                               cfg=cfg,
                               result_dir=str(RES_DATA_DIR + '/' + sequence + '-eval/'),
                               prefix='',
                               save_plot=True,
                               show_plot=False,
                               save_statistics=True,
                               plot_timestamps=True,
                               plot_ranges=True,
                               plot_ranges_sorted=True,
                               plot_error=True,
                               plot_histogram=True,
                               verbose=True)


    @staticmethod
    def test_range_gen_outliers():
        sequence = 'T1_A3_loiter_2m_2023-08-31-20-58-20'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        topic_list=['/d01/ranging', '/a01/ranging', '/a02/ranging', '/a03/ranging']
        fn_meas_ranges = str(RES_DATA_DIR + '/' + sequence + '-all-true-ranges-stddev0-outliers0_1.csv')
        fn_gt_ranges = str(RES_DATA_DIR + '/' + sequence + '-all-true-ranges.csv')


        bagfile_out_noisy = str(RES_DATA_DIR + '/' + sequence + '-true-ranges-stddev0-outliers0_1.bag')

        # 1) create a dirty bag file
        if not os.path.isfile(bagfile_out_noisy):
            res = ROSbag_TrueRanges.extract(bagfile_in_name=bagfile_in,
                                            bagfile_out_name=bagfile_out_noisy,
                                            topic_pose='/d01/mavros/vision_pose/pose',
                                            cfg=cfg_file,
                                            verbose=True,
                                            stddev_range=0.001,
                                            perc_outliers=0.1,
                                            stddev_outlier=1.0)

        if not os.path.isfile(fn_meas_ranges):
            # 2) extract all measurements from the dirty bagfile
            res = TWR_ROSbag2CSV.extract_to_one(bagfile_name=bagfile_out_noisy,
                                                    topic_list=topic_list,
                                                    fn=fn_meas_ranges,
                                                    result_dir=RES_DATA_DIR,
                                                    verbose=True)

        bagfile_out = str(RES_DATA_DIR + '/' + sequence + '-true-ranges.bag')

        # 3) create a clean bag file
        if not os.path.isfile(bagfile_out):
            res = ROSbag_TrueRanges.extract(bagfile_in_name=bagfile_in,
                                            bagfile_out_name=bagfile_out,
                                            topic_pose='/d01/mavros/vision_pose/pose',
                                            cfg=cfg_file,
                                            verbose=True,
                                            stddev_range=0.001)

        if not os.path.isfile(fn_gt_ranges):
            # 4) extract all measurements from the clean bagfile
            res = TWR_ROSbag2CSV.extract_to_one(bagfile_name=bagfile_out,
                                                    topic_list=topic_list,
                                                    fn=fn_gt_ranges,
                                                    result_dir=RES_DATA_DIR,
                                                    verbose=True)

        # 4) evaluate the ranges
        cfg = AssociateRangesCfg(ID1=None,
                                 ID2=None,
                                 relative_timestamps=False,
                                 max_difference=0.03,
                                 subsample=0,
                                 verbose=True,
                                 remove_outliers=False,
                                 max_range=30,
                                 range_error_val=0,
                                 label_timestamp='t',
                                 label_ID1='UWB_ID1',
                                 label_ID2='UWB_ID2',
                                 label_range='range_raw')

        eval = RangeEvaluation(fn_gt=fn_gt_ranges,
                               fn_est=fn_meas_ranges,
                               UWB_ID1_arr=[0,1,2,3],
                               UWB_ID2_arr=[0,1,2,3],
                               cfg=cfg,
                               result_dir=str(RES_DATA_DIR + '/' + sequence + '-eval-outliers/'),
                               prefix='',
                               save_plot=True,
                               show_plot=True,
                               save_statistics=True,
                               plot_timestamps=True,
                               plot_ranges=True,
                               plot_ranges_sorted=True,
                               plot_error=True,
                               plot_histogram=True,
                               verbose=True)


if __name__ == "__main__":
     RangeEvaluationTrail.test_range_gen_outliers()


#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# os
########################################################################################################################
import os
from sys import version_info
from datetime import datetime
import argparse
import time
import math
import yaml
import pandas as pandas
import numpy as np

import cnspy_numpy_utils.numpy_statistics
from cnspy_timestamp_association.TimestampAssociation import TimestampAssociation
from cnspy_ranging_evaluation.AssociateRanges import AssociateRanges, AssociateRangesCfg
from matplotlib import pyplot as plt

class RangeEvaluation:

    report = None
    def __init__(self,
                 fn_gt,
                 fn_est,
                 UWB_ID1_arr=[0], UWB_ID2_arr=[1],
                 result_dir=None,
                 prefix=None,
                 save_plot=True,
                 save_statistics=True,
                 show_plot=True,
                 cfg = AssociateRangesCfg(),
                 plot_timestamps=True,
                 plot_ranges=True,
                 plot_ranges_sorted=True,
                 plot_error =True,
                 plot_histogram=True,
                 verbose=False,
                 filter_histogram=True
                 ):
        if not result_dir:
            result_dir = '.'
        if not prefix:
            prefix = ''

        statistics_file = None
        if save_statistics and not plot_histogram:
            print("RangeEvaluation: Warning save_statistics can only be used in combination with plot_histogram")
            plot_histogram = True
        if save_statistics:
            if not os.path.exists(result_dir):
                os.makedirs(result_dir)
            stat_fn = os.path.join(result_dir, 'statistics.yaml')
            statistics_file = open(stat_fn, 'w')
            if verbose:
                print("RangeEvaluation: stat_fn=" + stat_fn)
            yaml.dump({'info': 'RangeEvaluation Statistics',
                       'time': datetime.now().strftime("%m/%d/%Y, %H:%M:%S"),
                       'fn_gt': fn_gt,
                       'fn_est': fn_est,
                       'filter_histogram': filter_histogram,
                       'result_dir': result_dir}, statistics_file, sort_keys=False, explicit_start=False, default_flow_style=False)
            yaml.dump({'UWB_ID1s': UWB_ID1_arr,
                       'UWB_ID2s': UWB_ID2_arr}, statistics_file, sort_keys=False, explicit_start=False, default_flow_style=True)
        fn_gt = os.path.abspath(fn_gt)
        fn_est = os.path.abspath(fn_est)

        plt.style.use('ggplot')

        SMALL_SIZE = 6
        MEDIUM_SIZE = 7
        BIGGER_SIZE = 8

        plt.rc('font', size=SMALL_SIZE)  # controls default text sizes
        plt.rc('axes', titlesize=SMALL_SIZE)  # fontsize of the axes title
        plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
        plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

        for UWB_ID1 in UWB_ID1_arr:
            if plot_timestamps:
                fig_t = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_t.suptitle('Timestamps of ID=' + str(UWB_ID1), fontsize=16)
            if plot_ranges:
                fig_r = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_r.suptitle('Ranges of ID=' + str(UWB_ID1), fontsize=16)
            if plot_ranges_sorted:
                fig_rs = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_rs.suptitle('Range sorted of ID=' + str(UWB_ID1), fontsize=16)
            if plot_error:
                fig_e = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_e.suptitle('Error of ID=' + str(UWB_ID1), fontsize=16)
            if plot_histogram:
                fig_h = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_h.suptitle('Histograms of ID=' + str(UWB_ID1), fontsize=16)
            if save_statistics:
                dict_statistics_i = {'ID' : UWB_ID1, 'contant_bias_table' : dict(), 'noise_table' : dict()}
                pass

            n = len(UWB_ID2_arr)
            sqrt_n = math.floor(math.sqrt(n))
            n_rows = sqrt_n
            if sqrt_n*sqrt_n < n:
                n_cols = sqrt_n+1
            else:
                n_cols = sqrt_n

            idx = 1
            for UWB_ID2 in UWB_ID2_arr:
                if UWB_ID1 == UWB_ID2:
                    continue

                cfg.ID1 = int(UWB_ID1)
                cfg.ID2 = int(UWB_ID2)
                cfg_title = str("ID" + str(UWB_ID1) + " to ID" + str(UWB_ID2))
                assoc = AssociateRanges(fn_gt=fn_gt, fn_est=fn_est, cfg=cfg)
                assoc.save(result_dir=result_dir, prefix=prefix+cfg_title)

                if plot_timestamps:
                    ax_t = fig_t.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_timestamps(fig=fig_t, ax=ax_t, calc_error=True, cfg_title=cfg_title)

                if plot_ranges:
                    ax_r = fig_r.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_ranges(fig=fig_r, ax=ax_r, cfg_title=cfg_title)

                if plot_ranges_sorted:
                    ax_rs = fig_rs.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_ranges(fig=fig_rs, ax=ax_rs, sorted=True, cfg_title=cfg_title)
                if plot_error:
                    ax_e = fig_e.add_subplot(n_rows, n_cols, idx)
                    [fig_, ax_, stat, r_vec_err_] = assoc.plot_range_error(fig=fig_e, ax=ax_e,
                                                                        sorted=False,
                                                                        remove_outlier=True,
                                                                        cfg_title=cfg_title)
                    cnspy_numpy_utils.numpy_statistics.print_statistics(stat, desc=cfg_title + " error")
                if plot_histogram:
                    ax_h = fig_h.add_subplot(n_rows, n_cols, idx)
                    [fig_, ax_, stat, r_vec_err_] = assoc.plot_range_error_histogram(fig=fig_h,
                                                                                     ax=ax_h,
                                                                                     max_error=1,
                                                                                     filter_histogramm=filter_histogram)
                    dict_statistics_i['contant_bias_table'][UWB_ID2] = round(float(stat['mean']),2)
                    dict_statistics_i['noise_table'][UWB_ID2] = round(float(stat['std']),2)

                # the histogram of the date
                idx += 1
            if verbose:
                print("* RangeEvaluation(): ranges associated!")

            # Tweak spacing to prevent clipping of ylabel
            if plot_timestamps:
                fig_t.tight_layout()
                if save_plot:
                    AssociateRanges.show_save_figure(fig=fig_t, result_dir=result_dir,
                                                     save_fn=str("Timestamps" + str(UWB_ID1)),
                                                     show=show_plot, close_figure=not show_plot)

            if plot_ranges:
               fig_r.tight_layout()
               if save_plot:
                   AssociateRanges.show_save_figure(fig=fig_r, result_dir=result_dir,
                                                    save_fn=str("Ranges_ID" + str(UWB_ID1)),
                                                    show=show_plot, close_figure=not show_plot)

            if plot_ranges_sorted:
                fig_rs.tight_layout()
                if save_plot:
                    AssociateRanges.show_save_figure(fig=fig_rs, result_dir=result_dir,
                                                     save_fn=str("Range_Sorted_ID" + str(UWB_ID1)),
                                                     show=show_plot, close_figure=not show_plot)

            if plot_error:
                fig_e.tight_layout()
                if save_plot:
                    AssociateRanges.show_save_figure(fig=fig_e, result_dir=result_dir,
                                                     save_fn=str("Errors_ID" + str(UWB_ID1)),
                                                     show=show_plot, close_figure=not show_plot)

            if plot_histogram:
                fig_h.tight_layout()
                if save_plot:
                    AssociateRanges.show_save_figure(fig=fig_h, result_dir=result_dir,
                                                     save_fn=str("Histograms_ID" + str(UWB_ID1)),
                                                     show=show_plot, close_figure=not show_plot)

            if save_statistics:
                yaml.dump(dict_statistics_i, statistics_file,explicit_start=True, default_flow_style=True)
                if verbose:
                    print(yaml.dump(dict_statistics_i,explicit_start=True, default_flow_style=True))
            pass # UWB_ID2
        pass # UWB_ID1

        if save_statistics:
            statistics_file.close()

        pass # DONE

def main():
    parser = argparse.ArgumentParser(
        description='RangeEvaluation: evaluate and estimated and true pairwise ranges')
    parser.add_argument('--fn_gt', help='input ground-truth trajectory CSV file', default="not specified")
    parser.add_argument('--fn_est', help='input estimated  trajectory CSV file', default="not specified")
    parser.add_argument('--result_dir', help='directory to store results [otherwise bagfile name will be a directory]',
                        default='')
    parser.add_argument('--UWB_ID1s', help='ID of TX', nargs='+', default=[0])
    parser.add_argument('--UWB_ID2s', help='ID of RX', nargs='+', default=[1])
    parser.add_argument('--prefix', help='prefix in results', default='')
    parser.add_argument('--max_timestamp_difference', help='Max difference between associated timestampes (t_gt - t_est)', default=0.03)
    parser.add_argument('--subsample', help='subsampling factor for input data (CSV)', default=0)
    parser.add_argument('--plot', action='store_true', default=True)
    parser.add_argument('--save_plot', action='store_true', default=True)
    parser.add_argument('--save_statistics', action='store_true', default=True)
    parser.add_argument('--show_plot', action='store_true', default=True)
    parser.add_argument('--relative_timestamps', action='store_true', default=False)
    parser.add_argument('--remove_outliers', action='store_true', default=False)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--max_range', help='range that classifies as outlier', default='30')
    parser.add_argument('--range_error_val', help='value assigned to outlier', default='0')
    parser.add_argument('--label_timestamp', help='timestamp label in CSV', default='t')
    parser.add_argument('--label_range', help='range label in CSV', default='range_raw')
    parser.add_argument('--label_ID1', help='ID1 label in CSV', default='UWB_ID1')
    parser.add_argument('--label_ID2', help='ID2 label in CSV', default='UWB_ID2')
    parser.add_argument('--plot_timestamps', action='store_true', default=False)
    parser.add_argument('--plot_ranges', action='store_true', default=False)
    parser.add_argument('--plot_ranges_sorted', action='store_true', default=False)
    parser.add_argument('--plot_errors', action='store_true', default=False)
    parser.add_argument('--plot_histograms', action='store_true', default=False)
    tp_start = time.time()
    args = parser.parse_args()
    cfg = AssociateRangesCfg(ID1=None,
                             ID2=None,
                             relative_timestamps=args.relative_timestamps,
                             max_difference=float(args.max_timestamp_difference),
                             subsample=int(args.subsample),
                             verbose=args.verbose,
                             remove_outliers=args.remove_outliers,
                             max_range=float(args.max_range),
                             range_error_val=float(args.range_error_val),
                             label_timestamp=args.label_timestamp,
                             label_ID1=args.label_ID1,
                             label_ID2=args.label_ID2,
                             label_range = args.label_range)

    eval = RangeEvaluation( fn_gt=args.fn_gt,
                            fn_est=args.fn_est,
                            UWB_ID1_arr=args.UWB_ID1s,
                            UWB_ID2_arr=args.UWB_ID2s,
                            cfg=cfg,
                            result_dir=args.result_dir,
                            prefix=args.prefix,
                            save_plot=args.save_plot,
                            show_plot=args.show_plot,
                            save_statistics=args.save_statistics,
                            plot_timestamps=args.plot_timestamps,
                            plot_ranges=args.plot_ranges,
                            plot_ranges_sorted=args.plot_ranges_sorted,
                            plot_error=args.plot_errors,
                            plot_histogram=args.plot_histograms,
                            )

    print(" ")
    print("finished after [%s sec]\n" % str(time.time() - tp_start))
    pass

if __name__ == "__main__":
    main()
    pass


#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# numpy, matplotlib
########################################################################################################################
import os
from sys import version_info
import time
import numpy as np
from matplotlib import pyplot as plt
import pandas as pandas
import math

from cnspy_numpy_utils.numpy_statistics import *
from cnspy_timestamp_association.TimestampAssociation import TimestampAssociation
from cnspy_trajectory.PlotLineStyle import PlotLineStyle


class AssociateRangesCfg:
    ID1 = None
    ID2 = None
    relative_timestamps = False
    max_difference = 0.02
    subsample = 0
    verbose = False
    remove_outliers = False,
    max_range = 30
    max_range_error = 10
    range_error_val = 0
    label_timestamp = 't'
    label_ID1 = 'ID1'
    label_ID2 = 'ID2'
    label_range = 'range_raw'
    def __init__(self, ID1=None, ID2=None, relative_timestamps=False,
                 max_difference=0.02, subsample=0, verbose=False, remove_outliers=False,
                 max_range=30, range_error_val=0, label_timestamp='t',
                 label_ID1='UWB_ID1',
                 label_ID2='UWB_ID2', label_range = 'range_raw'):
        self.label_timestamp = label_timestamp
        self.label_ID1 = label_ID1
        self.label_ID2 = label_ID2
        self.label_range = label_range
        self.ID1 = ID1
        self.ID2 = ID2
        self.relative_timestamps = relative_timestamps
        self.max_difference = max_difference
        self.subsample = subsample
        self.verbose = verbose
        self.remove_outliers = remove_outliers
        self.max_range = max_range
        self.range_error_val = range_error_val
        
        
    
class AssociateRanges:
    csv_df_gt = None
    csv_df_est = None

    data_frame_gt_matched = None
    data_frame_est_matched = None

    matches_est_gt = None  # list of tuples containing [(idx_est, idx_gt), ...]

    cfg = None
    data_loaded = False

    def __init__(self):
        pass

    def __init__(self, fn_gt, fn_est, cfg):
        assert (isinstance(cfg, AssociateRangesCfg))
        self.cfg = cfg
        self.load(fn_gt, fn_est, cfg)

    def load(self, fn_gt, fn_est, cfg):
        assert (isinstance(cfg, AssociateRangesCfg))
        assert (os.path.exists(fn_gt)), str("Path to fn_gt does not exist!:" + str(fn_gt))
        assert (os.path.exists((fn_est))), str("Path to fn_est does not exist!:" + str(fn_est))

        self.csv_df_gt = pandas.read_csv(fn_gt, sep='\s+|\,', comment='#', engine='python')
        self.csv_df_est = pandas.read_csv(fn_est, sep='\s+|\,', comment='#', engine='python')

        if cfg.ID1 is not None:
            self.csv_df_gt = (self.csv_df_gt.loc[self.csv_df_gt[cfg.label_ID1] == cfg.ID1])
            self.csv_df_est = (self.csv_df_est.loc[self.csv_df_est[cfg.label_ID1] == cfg.ID1])

            if len(self.csv_df_gt) == 0 or len(self.csv_df_est) == 0:
                print("ID1=" + str(cfg.ID1) + " not found")
                return

        if cfg.ID2 is not None:
            self.csv_df_gt = (self.csv_df_gt.loc[self.csv_df_gt[cfg.label_ID2] == cfg.ID2])
            self.csv_df_est = (self.csv_df_est.loc[self.csv_df_est[cfg.label_ID2] == cfg.ID2])
            if len(self.csv_df_gt) == 0 or len(self.csv_df_est) == 0:
                print("ID2=" + str(cfg.ID2) + " not found")
                return

        if cfg.remove_outliers:
            indices1 = np.nonzero(self.csv_df_est[cfg.label_range] < self.cfg.max_range)
            indices2 = np.nonzero(self.csv_df_est[cfg.label_range] > 0)
            idc = np.intersect1d(indices1, indices2)
            num_outliers = len(self.csv_df_est.index) - len(idc)
            perc = 100.0*(num_outliers / max(1, len(self.csv_df_est.index)))
            if num_outliers:
                self.csv_df_est = AssociateRanges.sample_DataFrame(self.csv_df_est, np.intersect1d(indices1, indices2))
                print('AssociateRanges.load(): [%d] outliers (%.1f %%) removed!' % (num_outliers, perc))

        else:
            indices = ((self.csv_df_est[cfg.label_range]) < 0)
            self.csv_df_est.loc[indices, cfg.label_range] = cfg.range_error_val
            indices = (self.csv_df_est[cfg.label_range] > cfg.max_range)
            self.csv_df_est.loc[indices, cfg.label_range] = cfg.range_error_val

        if cfg.subsample > 1:
            subsample = round(cfg.subsample, 0)
            self.csv_df_gt = AssociateRanges.subsample_DataFrame(df=self.csv_df_gt, step=subsample, verbose=cfg.verbose)
            self.csv_df_est = AssociateRanges.subsample_DataFrame(df=self.csv_df_est, step=subsample, verbose=cfg.verbose)

        if version_info[0] < 3:
            t_vec_gt = self.csv_df_gt.as_matrix([cfg.label_timestamp])
            t_vec_est = self.csv_df_est.as_matrix([cfg.label_timestamp])
            t_zero = min(t_vec_gt[0], t_vec_est[0])

            if len(t_vec_gt) == 0 or len(t_vec_est) == 0:
                print("empty")
                return

            if cfg.relative_timestamps:
                self.csv_df_gt[[cfg.label_timestamp]] = self.csv_df_gt[[cfg.label_timestamp]] - t_zero
                self.csv_df_est[[cfg.label_timestamp]] = self.csv_df_est[[cfg.label_timestamp]] - t_zero
        else:
            # FIX(scm): for newer versions as_matrix is deprecated, using to_numpy instead
            # from https://stackoverflow.com/questions/60164560/attributeerror-series-object-has-no-attribute-as-matrix-why-is-it-error
            t_vec_gt = self.csv_df_gt[[cfg.label_timestamp]].to_numpy()
            t_vec_est = self.csv_df_est[[cfg.label_timestamp]].to_numpy()
            if len(t_vec_gt) == 0 or len(t_vec_est) == 0:
                print("empty")
                return

            t_zero = min(t_vec_gt[0], t_vec_est[0])
            if cfg.relative_timestamps:
                self.csv_df_gt[[cfg.label_timestamp]] = self.csv_df_gt[[cfg.label_timestamp]] - t_zero
                self.csv_df_est[[cfg.label_timestamp]] = self.csv_df_est[[cfg.label_timestamp]] - t_zero

        if cfg.relative_timestamps:
            # only relative time stamps:
            t_vec_gt = t_vec_gt - t_zero
            t_vec_est = t_vec_est - t_zero

        idx_est, idx_gt, t_est_matched, t_gt_matched = TimestampAssociation.associate_timestamps(
            t_vec_est,
            t_vec_gt,
            max_difference=cfg.max_difference,
            round_decimals=6,
            unique_timestamps=True)

        self.data_frame_est_matched = self.csv_df_est.iloc[idx_est, :]
        self.data_frame_gt_matched = self.csv_df_gt.iloc[idx_gt, :]
        self.matches_est_gt = zip(idx_est, idx_gt)

        if cfg.verbose:
            print("AssociateRanges(): {} timestamps associated.".format(len(idx_est)))

        self.data_loaded = True
        # using zip() and * operator to
        # perform Unzipping
        # res = list(zip(*test_list))

    def plot_timestamps(self, cfg_dpi=200, cfg_title="timestamps", calc_error=False, fig=None, ax=None,
                        colors=['r', 'g'], labels=['gt', 'est'],
                        ls_vec=[PlotLineStyle(linestyle='-'), PlotLineStyle(linestyle='-.')],
                        save_fn="", result_dir="."):
        if not self.data_loaded:
            return
        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg_dpi))
        if ax is None:
            ax = fig.add_subplot(111)
        if cfg_title:
            ax.set_title(cfg_title)

        if version_info[0] < 3:
            t_vec_gt = self.data_frame_gt_matched.as_matrix([self.cfg.label_timestamp])
            t_vec_est = self.data_frame_est_matched.as_matrix([self.cfg.label_timestamp])
        else:
            t_vec_gt = self.data_frame_gt_matched[[self.cfg.label_timestamp]].to_numpy()
            t_vec_est = self.data_frame_est_matched[[self.cfg.label_timestamp]].to_numpy()

        if not calc_error:
            x_arr = range(len(t_vec_gt))
            AssociateRanges.ax_plot_n_dim(ax, x_arr, t_vec_gt, colors=[colors[0]], labels=[labels[0]], ls=ls_vec[0])
            x_arr = range(len(t_vec_est))
            AssociateRanges.ax_plot_n_dim(ax, x_arr, t_vec_est, colors=[colors[1]], labels=[labels[1]], ls=ls_vec[1])

            ax.grid(both=True)
            ax.set_xlabel('idx')
            ax.set_ylabel('time [s]')
        else:
            x_arr = range(len(t_vec_gt))
            # t_est = t_true + t_err
            t_vec_err = t_vec_est - t_vec_gt
            AssociateRanges.ax_plot_n_dim(ax, x_arr, t_vec_err, colors=['r'], labels=['err'], ls=ls_vec[0])
            ax.grid()
            ax.set_xlabel('idx')
            ax.set_ylabel('diff time [s]')

        AssociateRanges.show_save_figure(fig=fig, result_dir=result_dir, save_fn=save_fn, show=False)

        ax.legend()
        return fig, ax

    def plot_ranges(self, cfg_dpi=200, cfg_title="ranges", sorted=False, fig=None, ax=None,
                    colors=['r', 'g'], labels=['gt', 'est'],
                    ls_vec=[PlotLineStyle(linestyle='-'), PlotLineStyle(linestyle='-.')],
                    save_fn="", result_dir="."):
        if not self.data_loaded:
            return
        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg_dpi))
        if ax is None:
            ax = fig.add_subplot(111)
        if cfg_title:
            ax.set_title(cfg_title)

        if version_info[0] < 3:
            t_vec_gt = self.data_frame_gt_matched.as_matrix([self.cfg.label_timestamp])
            t_vec_est = self.data_frame_est_matched.as_matrix([self.cfg.label_timestamp])
        else:
            t_vec_gt = self.data_frame_gt_matched[[self.cfg.label_timestamp]].to_numpy()
            t_vec_est = self.data_frame_est_matched[[self.cfg.label_timestamp]].to_numpy()

        if version_info[0] < 3:
            r_vec_gt = self.data_frame_gt_matched.as_matrix([self.cfg.label_range])
            r_vec_est = self.data_frame_est_matched.as_matrix([self.cfg.label_range])
        else:
            r_vec_gt = self.data_frame_gt_matched[[self.cfg.label_range]].to_numpy()
            r_vec_est = self.data_frame_est_matched[[self.cfg.label_range]].to_numpy()

        if not sorted:
            # x_arr = range(len(t_vec_gt))
            AssociateRanges.ax_plot_n_dim(ax, t_vec_gt, r_vec_gt, colors=[colors[0]], labels=[labels[0]], ls=ls_vec[0])
            # x_arr = range(len(t_vec_est))
            AssociateRanges.ax_plot_n_dim(ax, t_vec_est, r_vec_est, colors=[colors[1]], labels=[labels[1]], ls=ls_vec[1])

            ax.grid()
            ax.set_ylabel('range')
            ax.set_xlabel('time [s]')
            AssociateRanges.show_save_figure(fig=fig, result_dir=result_dir, save_fn=save_fn, show=False)

        else:
            gt_indices_sorted = np.argsort(r_vec_gt, axis=0)
            x_arr = range(len(r_vec_gt))
            AssociateRanges.ax_plot_n_dim(ax, x_arr, np.take_along_axis(r_vec_gt, gt_indices_sorted, axis=0), colors=[colors[0]], labels=[labels[0]], ls=ls_vec[0])
            AssociateRanges.ax_plot_n_dim(ax, x_arr, np.take_along_axis(r_vec_est, gt_indices_sorted, axis=0), colors=[colors[1]], labels=[labels[1]], ls=ls_vec[1])
            ax.grid()
            ax.set_ylabel('range')
            ax.set_xlabel('range sorted index')
            AssociateRanges.show_save_figure(fig=fig, result_dir=result_dir, save_fn=save_fn, show=False)

        ax.legend()
        return fig, ax

    def get_range_error(self, sort=False, remove_outlier=True, max_error=None):
        if not self.data_loaded:
            return
        if version_info[0] < 3:
            t_vec_gt = self.data_frame_gt_matched.as_matrix([self.cfg.label_timestamp])
            t_vec_est = self.data_frame_est_matched.as_matrix([self.cfg.label_timestamp])
        else:
            t_vec_gt = self.data_frame_gt_matched[[self.cfg.label_timestamp]].to_numpy()
            t_vec_est = self.data_frame_est_matched[[self.cfg.label_timestamp]].to_numpy()

        if version_info[0] < 3:
            r_vec_gt = self.data_frame_gt_matched.as_matrix([self.cfg.label_range])
            r_vec_est = self.data_frame_est_matched.as_matrix([self.cfg.label_range])
        else:
            r_vec_gt = self.data_frame_gt_matched[[self.cfg.label_range]].to_numpy()
            r_vec_est = self.data_frame_est_matched[[self.cfg.label_range]].to_numpy()

        if not sort:
            # if r_est = r_true + r_err, then  r_err is an offset or the constant bias (gamma).
            r_vec_err = r_vec_est - r_vec_gt
            t_vec = t_vec_gt

            if max_error:
                indices = np.nonzero((abs(r_vec_err) > max_error))
                t_vec = np.delete(t_vec, indices, axis=0, )
                r_vec_err = np.delete(r_vec_err, indices, axis=0)

            return [t_vec, r_vec_err]
        else:
            # if r_est = r_true + r_err, then  r_err is an offset or the constant bias (gamma).
            r_vec_err = r_vec_est - r_vec_gt
            x_arr = r_vec_gt

            if max_error:
                indices = np.nonzero((abs(r_vec_err) > max_error))
                x_arr = np.delete(x_arr, indices, axis=0, )
                r_vec_err = np.delete(r_vec_err, indices, axis=0)

            gt_indices_sorted = np.argsort(x_arr, axis=0)
            #x_arr = range(len(r_vec_gt))
            t_vec = np.take_along_axis(x_arr, gt_indices_sorted, axis=0)
            r_vec = np.take_along_axis(r_vec_err, gt_indices_sorted, axis=0)
            return [t_vec, r_vec]

    def plot_range_error(self, cfg_dpi=200, cfg_title="ranges", sorted=False, remove_outlier=True, fig=None, ax=None,
                    colors=['r'], labels=['error'],
                    ls_vec=[PlotLineStyle(linestyle='-'), PlotLineStyle(linestyle='-.')],
                    save_fn="", result_dir="."):
        if not self.data_loaded:
            return
        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg_dpi))
        if ax is None:
            ax = fig.add_subplot(111)
        if cfg_title:
            ax.set_title(cfg_title)

        [t_vec, r_vec_err] = self.get_range_error(sort=sorted, remove_outlier=remove_outlier)
        if not sorted:
            AssociateRanges.ax_plot_n_dim(ax, t_vec, r_vec_err, colors=[colors[0]], labels=[labels[0]], ls=ls_vec[0])
            ax.grid()
            ax.set_ylabel('range error (est-gt)')
            ax.set_xlabel('time [s]')
            AssociateRanges.show_save_figure(fig=fig, result_dir=result_dir, save_fn=save_fn, show=False)

        else:
            AssociateRanges.ax_plot_n_dim(ax, t_vec, r_vec_err,
                                          colors=[colors[0]], labels=[labels[0]], ls=ls_vec[0])

            ax.grid()
            ax.set_ylabel('range error (est-gt)')
            ax.set_xlabel('range sorted index')
            AssociateRanges.show_save_figure(fig=fig, result_dir=result_dir, save_fn=save_fn, show=False)

        ax.legend()
        stat = numpy_statistics(vNumpy=np.squeeze(np.asarray(r_vec_err)))
        return fig, ax, stat, r_vec_err

    def plot_range_error_histogram(self, cfg_dpi=200, fig=None, ax=None,
                                   save_fn="", result_dir=".", max_error=None, filter_histogramm=False, perc_inliers = 0.3,
                                   ID1=None, ID2=None):
        if not self.data_loaded:
            return fig, ax, None, None

        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg_dpi))
        if ax is None:
            ax = fig.add_subplot(111)

        if ID1 is None:
            ID1 = self.cfg.ID1
        if ID2 is None:
            ID2 = self.cfg.ID2

        [t_vec, r_vec_err] = self.get_range_error(sort=True, remove_outlier=True, max_error=max_error)
        num_bins = 50
        n, bins, patches = ax.hist(r_vec_err, num_bins, density=True, color='red', alpha=0.75, label='Histogram')

        if len(t_vec) == 0 or len(r_vec_err) == 0:
            return fig, ax, None, None

        if not filter_histogramm:
            # add a 'best fit' line
            stat = numpy_statistics(vNumpy=np.squeeze(np.asarray(r_vec_err)))
            sigma = stat['std']
            mu = stat['mean']
            y = ((1 / (np.sqrt(2 * np.pi) * sigma)) *
                 np.exp(-0.5 * (1 / sigma * (bins - mu)) ** 2))

            scaling = len(r_vec_err)/num_bins
            ax.plot(bins, y*scaling, '--', color='blue', alpha=0.75, label='PDF')
            ax.set_ylabel('num. samples normalized')
            ax.set_xlabel('error [m]')
            ax.set_title(r'Range Error Histogram ID' + str(ID1) + '-ID' + str(ID2) + ': $\mu$=' + str(round(mu, 3)) + ', $\sigma$=' + str(round(sigma, 3)))
            ax.legend()
            return fig, ax, stat, r_vec_err
        else:
            idx_n_sorted = np.argsort(n)

            # assuming a certain percentage as inliers:
            num_best_bins = int(num_bins*(0.5*perc_inliers))

            # compute the mean about the most frequent values:
            idx_best_bins = idx_n_sorted[-num_best_bins:]
            best_error_vals = bins[idx_best_bins]
            mean_best_errors = np.mean(best_error_vals, axis=0)

            # compute a boundary to remove outliers:
            min_offset_errors = mean_best_errors - np.min(best_error_vals, axis=0)
            max_offset_errors = np.max(best_error_vals, axis=0) - mean_best_errors
            print('mean_best_errors %f' % mean_best_errors)

            # remove outliers
            r_filtered_err = r_vec_err[(r_vec_err > (mean_best_errors - min_offset_errors*2.0)) &
                                       (r_vec_err < (mean_best_errors + 2.0*max_offset_errors))]

            # add a 'best fit' line
            stat = numpy_statistics(vNumpy=np.squeeze(np.asarray(r_filtered_err)))
            num_plot_bins = int(num_bins*(perc_inliers))
            n_, bins_, patches_ = ax.hist(r_filtered_err, num_plot_bins, density=True, color='blue', alpha=0.75, label='Histogram (filtered)')
            sigma = max(0.001, stat['std'])
            mu = stat['mean']
            scaling = 1.0; #len(r_filtered_err)/num_plot_bins
            y = ((1 / (np.sqrt(2 * np.pi) * sigma)) *
                 np.exp(-0.5 * (1 / sigma * (bins_ - mu)) ** 2))
            ax.plot(bins_, y*scaling, '--', color='green', label='PDF (filtered)')
            ax.set_ylabel('num. samples normalized')
            ax.set_xlabel('error [m]')
            ax.set_title(r'Range Error Histogram (filtered) ID' + str(ID1) + '-ID' + str(ID2) + ': $\mu$=' + str(round(mu, 3)) + ', $\sigma$=' + str(round(sigma, 3)))
            ax.legend()
            return fig, ax, stat, r_filtered_err
        pass

    def save(self, result_dir=None, prefix=None):
        if not self.data_loaded:
            return
        if not result_dir:
            fn_est_ = str(os.path.splitext(self.csv_df_est.fn)[0]) + "_matched.csv"
            fn_gt_ = str(os.path.splitext(self.csv_df_gt.fn)[0]) + "_matched.csv"
        else:
            if not prefix:
                prefix = ""
            if not os.path.exists(result_dir):
                os.makedirs(os.path.abspath(result_dir))
            fn_est_ = str(result_dir) + '/' + str(prefix) + "est_matched.csv"
            fn_gt_ = str(result_dir) + '/' + str(prefix) + "gt_matched.csv"

        self.data_frame_est_matched.to_csv(fn_est_, sep=',', index=False)
        self.data_frame_gt_matched.to_csv(fn_gt_, sep=',', index=False)



    @staticmethod
    def subsample_DataFrame(df, step=None, num_max_points=None, verbose=False):

        num_elems = len(df.index)

        if num_max_points:
            step = 1
            if (int(num_max_points) > 0) and (int(num_max_points) < num_elems):
                step = int(math.ceil(num_elems / float(num_max_points)))

        sparse_indices = np.arange(start=0, stop=num_elems, step=step)

        if num_max_points or step:
            if verbose:
                print("subsample_DataFrame():")
                print("* len: " + str(num_elems) + ", max_num_points: " + str(
                    num_max_points) + ", subsample by: " + str(step))

            return AssociateRanges.sample_DataFrame(df, sparse_indices)
        else:
            return df.copy(True)

    @staticmethod
    def sample_DataFrame(df, indices_arr):
        num_elems = len(df.index)
        assert (len(indices_arr) <= num_elems), "sample_DataFrame():\n\t index array must be smaller " \
                                                "equal the dataframe."
        assert (max(indices_arr) <= num_elems), "sample_DataFrame():\n\t elements in the index array " \
                                                "must be smaller equal the dataframe."
        assert (min(indices_arr) >= 0), "sample_DataFrame():\n\t elemts in the index array " \
                                        "must be greater equal zero."

        df_sub = df.iloc[indices_arr]
        df_sub.reset_index(inplace=True, drop=True)
        return df_sub

    @staticmethod
    def ax_plot_n_dim(ax, x_linespace, values,
                      colors=['r', 'g', 'b'],
                      labels=['x', 'y', 'z'], ls=PlotLineStyle()):
        assert len(colors) == len(labels)
        if len(colors) > 1:
            assert len(colors) == values.shape[1]
            for i in range(len(colors)):
                ax.plot(x_linespace, values[:, i],
                        color=colors[i], label=labels[i], linestyle=ls.linestyle, linewidth=ls.linewidth,
                        marker=ls.marker)
        else:
            ax.plot(x_linespace, values, color=colors[0], label=labels[0], linestyle=ls.linestyle,
                    linewidth=ls.linewidth)

    @staticmethod
    def ax_x_linespace(ax, ts=None, dist_vec=None, relative_time=True, plot_type=None, x_label_prefix=''):
        if plot_type == "plot_2D_over_dist":
            x_arr = dist_vec
            ax.set_xlabel(x_label_prefix + 'distance [m]')
        else:
            x_arr = ts
            if relative_time:
                x_arr = x_arr - x_arr[0]
                ax.set_xlabel(x_label_prefix + 'rel. time [sec]')
            else:
                ax.set_xlabel(x_label_prefix + 'time [sec]')
        return x_arr

    @staticmethod
    def show_save_figure(fig, save_fn="", result_dir=".", dpi=200, show=True, close_figure=False):
        assert (isinstance(fig, plt.Figure))

        plt.pause(0.1)
        # set current active again
        plt.figure(fig.number)
        # draw
        fig.canvas.draw_idle()
        plt.pause(0.1)
        time.sleep(0.1)
        if save_fn:
            if not os.path.exists(result_dir):
                os.makedirs(result_dir)

            filename = os.path.join(result_dir, save_fn) + ".png"

            # create directories for files
            print("save to file: " + filename)
            [root, ext] = os.path.splitext(filename)
            [head, tail] = os.path.split(root)
            try:  # else already exists
                os.makedirs(head)
            except:
                pass

            #plt.savefig(fig=fig, fname=filename, dpi=float(dpi), format="png")
            fig.savefig(filename, dpi=float(dpi), format='png')
            print("save to file done...")
        if show:
            plt.show()
        if close_figure:
            plt.close(fig)



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2024, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################

import os
import yaml
from cnspy_ranging_evaluation.RangeEvaluation import *
from cnspy_ranging_evaluation.TWR_ROSbag2CSV import *
from cnspy_ranging_evaluation.ROSBag_TrueRanges import *


class RangeEvaluationTool:
    @staticmethod
    def evaluate(bagfile_in, cfg, result_dir=None, verbose=True, show_plot=True, save_plot=True):
        if not os.path.isfile(bagfile_in):
            print("RangeEvaluationTool: could not find file: %s" % bagfile_in)
            return False


        ## create result dir:
        if result_dir == "" or result_dir is None:
            folder = str(bagfile_in).replace(".bag", "")
        else:
            folder = result_dir

        folder = os.path.abspath(folder)
        try:  # else already exists
            os.makedirs(folder)
        except:
            pass
        result_dir = folder
        if verbose:
            print("* result_dir: \t " + str(folder))

        dict_cfg = None
        with open(cfg, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "tag_topics" not in dict_cfg:
                print("[tag_topics] does not exist in fn=" + cfg)
                return False
            if "anchor_topics" not in dict_cfg:
                print("[anchor_topics] does not exist in fn=" + cfg)
                return False
            if "pose_topic" not in dict_cfg:
                print("[pose_topic] does not exist in fn=" + cfg)
                return False
            print("Read successful")

        UWB_ID1_arr = []
        topic_list = []
        for key, val in dict_cfg["tag_topics"].items():
            topic_list.append(val)
            UWB_ID1_arr.append(int(key))
        for key, val in dict_cfg["anchor_topics"].items():
            topic_list.append(val)
            UWB_ID1_arr.append(int(key))

        if verbose:
            print("* topic_list= " + str(topic_list))
            print("* UWB_ID1/2_arr= " + str(UWB_ID1_arr))

        #topic_list=['/d01/ranging', '/a01/ranging', '/a02/ranging', '/a03/ranging']
        fn_meas_ranges = str(result_dir + '/all-meas-ranges.csv')
        fn_gt_ranges = str(result_dir + '/all-true-ranges.csv')


        # 1) extract all measurements to CSV
        if not os.path.isfile(fn_meas_ranges):
            res = TWR_ROSbag2CSV.extract_to_one(bagfile_name=bagfile_in,
                                                topic_list=topic_list,
                                                fn=fn_meas_ranges,
                                                result_dir=result_dir,
                                                verbose=verbose)



        bagfile_out = str(result_dir + '/true-ranges.bag')

        # 2) create a clean bag file
        if not os.path.isfile(bagfile_out):
            res = ROSbag_TrueRanges.extract(bagfile_in_name=bagfile_in,
                                            bagfile_out_name=bagfile_out,
                                            topic_pose=dict_cfg["pose_topic"],
                                            cfg=cfg,
                                            verbose=verbose,
                                            stddev_range=0.001)

        if not os.path.isfile(fn_gt_ranges):
            # 3) extract all measurements from the clean bagfile
            res = TWR_ROSbag2CSV.extract_to_one(bagfile_name=bagfile_out,
                                                topic_list=topic_list,
                                                fn=fn_gt_ranges,
                                                result_dir=result_dir,
                                                verbose=verbose)


        # 4) evaluate the ranges
        cfg = AssociateRangesCfg(ID1=None,
                                 ID2=None,
                                 relative_timestamps=False,
                                 max_difference=0.03,
                                 subsample=0,
                                 verbose=True,
                                 remove_outliers=False,
                                 max_range=30,
                                 range_error_val=0,
                                 label_timestamp='t',
                                 label_ID1='UWB_ID1',
                                 label_ID2='UWB_ID2',
                                 label_range='range_raw')

        eval = RangeEvaluation(fn_gt=fn_gt_ranges,
                               fn_est=fn_meas_ranges,
                               UWB_ID1_arr=UWB_ID1_arr,
                               UWB_ID2_arr=UWB_ID1_arr,
                               cfg=cfg,
                               result_dir=str(result_dir + '/eval/'),
                               prefix='',
                               save_plot=save_plot,
                               show_plot=show_plot,
                               save_statistics=True,
                               plot_timestamps=True,
                               plot_ranges=True,
                               plot_ranges_sorted=True,
                               plot_error=True,
                               plot_histogram=True,
                               verbose=verbose
                               )
        pass  # DONE




def main():
    # RangeEvaluationTool.py --bagfile  ./test/sample_data/T1_A3_loiter_2m_2023-08-31-20-58-20.bag --cfg ./test/sample_data/config.yaml
    parser = argparse.ArgumentParser(
        description='RangeEvaluationTool: evaluation the measured ranges')
    parser.add_argument('--result_dir', help='directory to store results [otherwise bagfile name will be a directory]',
                        default='')
    parser.add_argument('--bagfile', help='input bag file', default="not specified")
    parser.add_argument('--cfg',
                        help='YAML configuration file describing the setup: {rel_tag_positions, abs_anchor_positions}',
                        default="config.yaml", required=True)
    parser.add_argument('--save_plot', action='store_true', default=True)
    parser.add_argument('--show_plot', action='store_true', default=False)
    parser.add_argument('--verbose', action='store_true', default=True)

    tp_start = time.time()
    args = parser.parse_args()

    RangeEvaluationTool.evaluate(bagfile_in=args.bagfile,
                                 cfg=args.cfg,
                                 result_dir=args.result_dir,
                                 save_plot=args.save_plot,
                                 show_plot=args.show_plot,
                                 verbose=args.verbose)
    pass
    print(" ")
    print("finished after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass



#!/usr/bin/env python

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cnspy_ranging_evaluation',
    version="0.2.8",
    author='Roland Jung',
    author_email='roland.jung@aau.at',
    description='Evaluation of range measurements.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/aau-cns/cnspy_ranging_evaluation/',
    project_urls={
        "Bug Tracker": "https://github.com/aau-cns/cnspy_ranging_evaluation/issues",
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],

    packages=find_packages(exclude=["test_*", "TODO*"]),
    python_requires='>=3.6',
    install_requires=['numpy', 'pandas', 'spatialmath-python', 'scipy', 'matplotlib', 'joblib', 'configparser', 'cnspy_numpy_utils', 'cnspy_timestamp_association' ],
    entry_points={
        'console_scripts': [
            'CSV_StaticBiasAnalysis = cnspy_ranging_evaluation.CSV_StaticBiasAnalysis:main',
            'IMU_ROSbag2CSV = cnspy_ranging_evaluation.IMU_ROSbag2CSV:main',
            'RangeEvaluationTool = cnspy_ranging_evaluation.RangeEvaluationTool:main',
            'ROSBag_Pose2Ranges = cnspy_ranging_evaluation.ROSBag_Pose2Ranges:main',
            'ROSBag_TrueRanges = cnspy_ranging_evaluation.ROSBag_TrueRanges:main',
            'TWR_ROSbag2CSV = cnspy_ranging_evaluation.TWR_ROSbag2CSV:main',
        ],
    },
)



